from typing import Union, List, Optional

from . import _groups
from ._ldaps import LdapObject
from ._object_classes import User, Group


def search(
    connection: LdapObject, search_base: Optional[str] = None, search_scope: Optional[int] = None, **search_params
) -> List[Group]:
    """ Search for an Active Directory group.

    :param connection: An LDAPS connection.
    :param search_base: An optional distinguishedName to use as the root of the search. Defaults
        to the root of the domain.
    :param search_scope: An optional int to use as the search scope. Defaults to 2 (SCOPE_SUBTREE).
    :param search_params: Keyword arguments to include in the LDAP query.

    :raises exceptions.InvalidFilter: in the unlikely case that search parameters, once automatically escaped,
        result in an invalid LDAP query.
    :raises exceptions.UserAccountDisabled: if the user account performing the operation is disabled.

    :returns: A list of Active Directory group objects.

    To find the group with cn 'test.group', call the function like so:

        .. code-block:: python

            groups.search(
                connection,
                cn="test.group"
            )

    To find all groups with a name starting with 'sales', call the function like so:

        .. code-block:: python

            groups.search(connection, name="sales*")

    .. note:: If the wildcard character is not at the end of the search string, it will be escaped.

    """
    return _groups._get(connection, search_base=search_base, **search_params, search_scope=search_scope)


def get_all(
    connection: LdapObject, search_base: Optional[str] = None, search_scope: Optional[int] = None
) -> List[Group]:
    """ Get all Active Directory groups.

    :param connection: An LDAPS connection.
    :param search_base: An optional distinguishedName to use as the root of the search. Defaults
        to the root of the domain.
    :param search_scope: An optional int to use as the search scope. Defaults to 2 (SCOPE_SUBTREE).

    :raises exceptions.UserAccountDisabled: if the user account performing the operation is disabled.

    :returns: A list of Active Directory group objects.

    Example:

        .. code-block:: python

            groups.get_all(connection)

    .. warning:: This operation can be expensive! Using the search function is recommended.

    .. seealso:: :ref:`active_directory:Search Functions`
    """
    return _groups._get(connection, search_base=search_base, search_scope=search_scope)


def group_memberships(
    connection: LdapObject, *, distinguished_name: str, nested: Optional[bool] = False
) -> List[Group]:
    """ List the group memberships of a group.

    :param connection: An LDAPS connection.
    :param distinguished_name: The distinguished name of the target group.
    :param nested: Optional. Defaults to False. If set to `True` the function will return both direct and
        nested group memberships.

    .. important:: Finding nested memberships can be very slow!

    :raises exceptions.NoSuchObject: if the group does not exist.
    :raises exceptions.UserAccountDisabled: if the user account performing the operation is disabled.

    :returns: A list of Active Directory group objects.

    .. seealso:: :ref:`active_directory:Search Functions`
    """
    return _groups._group_memberships(connection, distinguished_name, nested=nested)


def members(
    connection: LdapObject, *, distinguished_name: str, nested: Optional[bool] = False
) -> List[Union[Group, User]]:
    """ Get the members of a group and return a list of Active Directory
    user/group objects.

    :param connection: An LDAPS connection.
    :param distinguished_name: The distinguished name of the target group.
    :param nested: Optional. Defaults to False. If set to `True` the function will return both direct and
        nested group memberships.

    .. important:: Finding nested memberships can be very slow!

    Example:

        .. code-block:: python

            group_members = groups.members(
                connection,
                distinguishedName="cn=test group,cn=users,dc=test,dc=domain,dc=com"
            )
            users_in_group = [m for m in group_members if m._type == 'user']
            groups_in_group = [m for m in group_members if m._type == 'group']

    :raises exceptions.NoSuchObject: if the group does not exist.
    :raises exceptions.UserAccountDisabled: if the user account performing the operation is disabled.

    :returns: A list of Active Directory group/user objects. You can inspect the type
        of each returned member with the `_type` property.

    .. warning:: This operation is expensive! Use sparingly, and consider whether getting the group
        memberships of a user with `users.group_memberships` could help you instead.

    .. seealso:: :ref:`active_directory:Search Functions`
    """
    return _groups._members(connection, distinguished_name, nested=nested)


def member_of(connection: LdapObject, *, distinguished_name: str) -> List[Group]:
    return _groups._member_of(connection, distinguished_name)


def add_member(connection: LdapObject, *, distinguished_name: str, member_distinguished_name: str) -> None:
    """ Add a member to a group.

    :param connection: An LDAPS connection.
    :param distinguished_name: The distinguished name of the target group.
    :param member_distinguished_name: The distinguished name of the member to add.

    :raises exceptions.NoSuchObject: if either the supplied group or member does not exist.
    :raises exceptions.ObjectExists: if the supplied member is already in the target group.
    :raises exceptions.UserAccountDisabled: if the user account performing the operation is disabled.

    Example:

        .. code-block:: python

            groups.add_member(
                connection,
                distinguished_name='cn=test group,cn=users,dc=test,dc=domain,dc=com',
                member_distinguished_name='cn=test user,cn=users,dc=test,dc=domain,dc=com',
            )

    .. seealso:: :ref:`active_directory:Action Functions`
    """
    return _groups._modify_members(connection, distinguished_name, member_distinguished_name, "add")


def remove_member(connection: LdapObject, *, distinguished_name: str, member_distinguished_name: str) -> None:
    """ Remove a member from a group.

    :param connection: An LDAPS connection.
    :param distinguished_name: The distinguished name of the target group.
    :param member_distinguished_name: The distinguished name of the member to remove.

    :raises exceptions.NoSuchObject: if either the supplied group or member does not exist.
    :raises exceptions.OperationFailed: if the supplied member is not in the target group.
    :raises exceptions.UserAccountDisabled: if the user account performing the operation is disabled.

    Example:

        .. code-block:: python

            groups.remove_member(
                connection,
                distinguished_name='cn=test group,cn=users,dc=test,dc=domain,dc=com',
                member_distinguished_name='cn=test user,cn=users,dc=test,dc=domain,dc=com',
            )

    .. seealso:: :ref:`active_directory:Action Functions`
    """
    return _groups._modify_members(connection, distinguished_name, member_distinguished_name, "delete")


def create(
    connection: LdapObject, *, distinguished_name: str, name: str, group_type: str = "global", **extra_params
) -> None:
    """ Create a new Active Directory security group.

    :param connection: An LDAPS connection.

    :param distinguished_name: The distinguished name of the target group.
    :param name: The name for the new group.
    :param group_type: The type of the group. Defaults to 'global'.
    :param extra_params: Keyword arguments to include in the new user LDAP attributes.

    :raises exceptions.ObjectExists: if the supplied user already exists.
    :raises exceptions.ParameterError:  if an invalid LDAP attribute or `group_type` is supplied.
    :raises exceptions.UserAccountDisabled: if the user account performing the operation is disabled.

    Example:

        .. code-block:: python

            groups.create(
                connection,
                distinguished_name='cn=test group,cn=Users,dc=test,dc=domain,dc=com',
                name='group name',
                group_type='security',
                info='my new group'
            )

    .. seealso:: :ref:`active_directory:Action Functions`, :ref:`active_directory:Keyword Argument Attributes`
    """
    return _groups._create(connection, distinguished_name, name, group_type, **extra_params)


def delete(connection: LdapObject, *, distinguished_name: str) -> None:
    """ Delete a group.

    :param connection: An LDAPS connection.
    :param distinguished_name: The distinguished name of the target group.

    :raises exceptions.NoSuchObject: if the group does not exist.
    :raises exceptions.UserAccountDisabled: if the user account performing the operation is disabled.

    Example:

        .. code-block:: python

            groups.delete(
                connection,
                distinguished_name='cn=my old group,cn=users,dc=test,dc=domain,dc=com',
            )

    .. seealso:: :ref:`active_directory:Action Functions`
    """
    return _groups._delete(connection, distinguished_name)


def move(connection: LdapObject, *, distinguished_name: str, new_parent: str) -> None:
    """ Move a group.

    :param connection: An LDAPS connection.
    :param distinguished_name: The distinguished name of the target group.
    :param new_parent: The distinguished name of the new OU or CN to put the group in.

    :raises exceptions.NoSuchObject: if the group or new parent does not exist.
    :raises exceptions.UserAccountDisabled: if the user account performing the operation is disabled.

    Example:

        .. code-block:: python

            groups.move(
                connection,
                distinguished_name='cn=engineers,cn=users,dc=test,dc=domain,dc=com',
                new_parent='ou=staff groups,dc=test,dc=domain,dc=com'
            )

    .. seealso:: :ref:`active_directory:Action Functions`
    """
    return _groups._move(connection, distinguished_name, new_parent)


def set_attribute(connection: LdapObject, *, distinguished_name: str, name: str, value: Union[int, str, bytes]) -> None:
    """ Set a group LDAP attribute value.

    This function performs the LDAP `modify_add` operation. It is suitable for setting
    values against blank attributes.

    If the supplied attribute already has a value, exceptions.ModifyAttributeError
    will be raised. In this case modify_attribute should be used instead.

    :param connection: An LDAPS connection.
    :param distinguished_name: The distinguished name of the target group.
    :param attribute_name: The name of the LDAP attribute to modify.
    :param value: The value to set on the LDAP attribute.
    :raises exceptions.UserAccountDisabled: if the user account performing the operation is disabled.

    :raises exceptions.ModifyAttributeError: if there was an issue setting the new attribute value.
    :raises exceptions.NoSuchObject: if the group does not exist.

    Example:

        .. code-block:: python

            groups.set_attribute(
                connection,
                distinguished_name="cn=test group,cn=Users,dc=test,dc=domain,dc=com",
                name="info",
                value="my new value"
            )

    .. seealso:: :ref:`active_directory:Action Functions`

    """
    _groups._modify(connection, distinguished_name, name, value, "add")


def replace_attribute(
    connection: LdapObject, *, distinguished_name: str, name: str, value: Union[int, str, bytes]
) -> None:
    """ Modify a group LDAP attribute value.

    This function performs the LDAP `modify_replace` operation.

    As suggested by the function name; if the supplied attribute already has a value, it will
    be replaced.

    :param connection: An LDAPS connection.
    :param distinguished_name: The distinguished name of the target group.
    :param attribute_name: The name of the LDAP attribute to modify.
    :param value: The value to replace the existing LDAP attribute value with.

    :raises exceptions.ModifyAttributeError: if there was an issue replacing the attribute value.
    :raises exceptions.NoSuchObject: if the group does not exist.
    :raises exceptions.UserAccountDisabled: if the user account performing the operation is disabled.

    Example:

        .. code-block:: python

            replace_attribute(
                connection,
                distinguished_name="cn=test group,cn=Users,dc=test,dc=domain,dc=com",
                name="info",
                value="my new value"
            )

    .. seealso:: :ref:`active_directory:Action Functions`
    """
    _groups._modify(connection, distinguished_name, name, value, "replace")


def delete_attribute(
    connection: LdapObject, *, distinguished_name: str, name: str, value: Union[int, str, bytes]
) -> None:
    """ Remove a group LDAP attribute value.

    This function performs the LDAP `modify_delete` operation.

    :param connection: An LDAPS connection.
    :param distinguished_name: The distinguished name of the target group.
    :param attribute_name: The name of the LDAP attribute to modify.
    :param value: The attribute value to delete.

    :raises exceptions.ModifyAttributeError: if there was an issue replacing the attribute value.
    :raises exceptions.NoSuchObject: if the group does not exist.
    :raises exceptions.UserAccountDisabled: if the user account performing the operation is disabled.

    Example:

        .. code-block:: python

            groups.delete_attribute(
                connection,
                distinguished_name="cn=test group,cn=Users,dc=test,dc=domain,dc=com",
                name="info",
                value="known existing value"
            )

    .. seealso:: :ref:`active_directory:Action Functions`
    """
    _groups._modify(connection, distinguished_name, name, value, "delete")
